import asyncio

_LOOP = None


def get_loop():
    """Return a default asyncio event loop."""
    global _LOOP

    if _LOOP is not None:
        return _LOOP

    _LOOP = asyncio.get_event_loop()
    return _LOOP



# def queryName(self, name, ip='', port=137, timeout=30):
#     """
#     Send a query on the network and hopes that if machine matching the *name* will reply with its IP address.
#
#     :param string ip: If the NBNSProtocol instance was instianted with broadcast=True, then this parameter can be an empty string. We will leave it to the OS to determine an appropriate broadcast address.
#                       If the NBNSProtocol instance was instianted with broadcast=False, then you should provide a target IP to send the query.
#     :param integer port: The NetBIOS-NS port (IANA standard defines this port to be 137). You should not touch this parameter unless you know what you are doing.
#     :param integer/float timeout: Number of seconds to wait for a reply, after which the method will return None
#     :return: A list of IP addresses in dotted notation (aaa.bbb.ccc.ddd). On timeout, returns None.
#     """
#     assert self.sock, 'Socket is already closed'
#
#     trn_id = random.randint(1, 0xFFFF)
#     data = self.prepareNameQuery(trn_id, name)
#     if self.broadcast and not ip:
#         ip = '<broadcast>'
#     elif not ip:
#         self.log.warning(
#             'queryName: ip parameter is empty. OS might not transmit this query to the network')
#
#     self.write(data, ip, port)
#
#     return self._pollForNetBIOSPacket(trn_id, timeout)
#
# def _pollForNetBIOSPacket(self, wait_trn_id, timeout):
#     end_time = time.time() - timeout
#     while True:
#         try:
#             _timeout = time.time() - end_time
#             if _timeout <= 0:
#                 return None
#
#             ready, _, _ = select.select([self.sock.fileno()], [], [], _timeout)
#             if not ready:
#                 return None
#
#             data, _ = self.sock.recvfrom(0xFFFF)
#             if len(data) == 0:
#                 raise NotConnectedError
#
#             trn_id, ret = self.decodePacket(data)
#
#             if trn_id == wait_trn_id:
#                 return ret
#         except select.error as ex:
#             if type(ex) is tuple:
#                 if ex[0] != errno.EINTR and ex[0] != errno.EAGAIN:
#                     raise ex
#             else:
#                 raise ex
#
#
# import asyncio
#
# class EchoClientProtocol:
#     def __init__(self, message, loop):
#         self.message = message
#         self.loop = loop
#         self.transport = None
#
#     def connection_made(self, transport):
#         self.transport = transport
#         print('Send:', self.message)
#         self.transport.sendto(self.message.encode())
#
#     def datagram_received(self, data, addr):
#         print("Received:", data.decode())
#
#         print("Close the socket")
#         self.transport.close()
#
#     def error_received(self, exc):
#         print('Error received:', exc)
#
#     def connection_lost(self, exc):
#         print("Socket closed, stop the event loop")
#         loop = asyncio.get_event_loop()
#         loop.stop()
#
# loop = asyncio.get_event_loop()
# message = "Hello World!"
# connect = loop.create_datagram_endpoint(
#     lambda: EchoClientProtocol(message, loop),
#     remote_addr=('127.0.0.1', 9999))
# transport, protocol = loop.run_until_complete(connect)
# loop.run_forever()
# transport.close()
# loop.close()


""" Example of browsing for a service (in this case, HTTP) """

import asyncio
import logging
import socket

from aiopvapi.helpers.aiorequest import AioRequest, PvApiResponseStatusError
from aiopvapi.hub import Hub
from zeroconf import Zeroconf, ServiceBrowser, ServiceStateChange

from pv_prompt.print_output import warn

LOGGER = logging.getLogger(__name__)


class Zero:
    def __init__(self, loop):
        self.hubs = []
        self._zeroconf = Zeroconf()
        self.browser = None
        self._loop = loop
        self._tasks = []

    async def add_update_task(self, hub):
        try:
            await hub.query_user_data()
        except PvApiResponseStatusError as err:
            warn('hub {} {} does not respond.'.format(hub.name,hub.ip))
            self.hubs.remove(hub)
        finally:
            await hub.request.websession.close()

    async def discover(self, wait_for=5):
        LOGGER.debug('starting discovery')
        self.browser = ServiceBrowser(
            self._zeroconf, "_powerview._tcp.local.",
            handlers=[self.on_service_state_change])

        LOGGER.debug("waiting for {} seconds".format(wait_for))
        await asyncio.sleep(wait_for)

        LOGGER.debug("closing discovery")
        self._zeroconf.close()

        LOGGER.debug("waiting for all tasks to finish.")
        try:
            await asyncio.gather(*self._tasks)
        except PvApiResponseStatusError as err:
            LOGGER.error(err)

    def on_service_state_change(self, zeroconf, service_type, name,
                                state_change):
        LOGGER.debug("Service %s of type %s state changed: %s" % (
            name, service_type, state_change))

        if state_change is ServiceStateChange.Added:
            info = zeroconf.get_service_info(service_type, name)
            if info:
                address = socket.inet_ntoa(info.address)
                LOGGER.debug("hub found with ip: {}".format(address))
                request = AioRequest(hub_ip=address, loop=self._loop)
                hub = Hub(request)
                self.hubs.append(hub)
                self._tasks.append(
                    self._loop.create_task(self.add_update_task(hub)))

# if __name__ == '__main__':
#     z = Zero()
#     z.discover()
#     # logging.basicConfig(level=logging.DEBUG)
#     # if len(sys.argv) > 1:
#     #     assert sys.argv[1:] == ['--debug']
#     #     logging.getLogger('zeroconf').setLevel(logging.DEBUG)
#     #
#     # zeroconf = Zeroconf()
#     # print("\nBrowsing services, press Ctrl-C to exit...\n")
#     # browser = ServiceBrowser(zeroconf, "_powerview._tcp.local.",
#     #                          handlers=[on_service_state_change])
#     #
#     # try:
#     #     sleep(5)
#     #     zeroconf.close()
#     # except KeyboardInterrupt:
#     #     pass
#     # finally:
#     #     zeroconf.close()



import asyncio
from aiopvapi.helpers.aiorequest import AioRequest, PvApiConnectionError
from aiopvapi.helpers.api_base import ApiEntryPoint
from aiopvapi.rooms import Rooms as PvRooms
from aiopvapi.scene_members import SceneMembers as PvSceneMembers
from aiopvapi.scenes import Scenes as PvScenes
from aiopvapi.shades import Shades as PvShades
from prompt_toolkit.completion import WordCompleter

from pv_prompt.base_prompts import BasePrompt, InvalidIdException
from pv_prompt.helpers import get_loop
from pv_prompt.print_output import print_waiting_done


class ResourceCache:
    def __init__(self,
                 api_entry_point: ApiEntryPoint,
                 resource_type_name: str,
                 request: AioRequest):
        self.request = request
        self.api_entry_point = api_entry_point
        self.resources = []
        self.id_suggestions = []
        self.resource_type_name = resource_type_name

    def __iter__(self):
        return iter(self.resources)

    def __len__(self):
        return len(self.resources)

    def get_name_by_id(self, _id):
        for _item in self.resources:
            if _item.id == _id:
                return _item.name

    def _populate_id_suggestions(self):
        self.id_suggestions = WordCompleter(
            [str(_item.id) for _item in self.resources])

    def find_by_id(self, id_: int):
        for _item in self.resources:
            if _item.id == id_:
                return _item
        raise InvalidIdException("No data found for id {}".format(id_))

    def _validate_id(self, _id):
        if _id is None:
            raise InvalidIdException
        try:
            _id = int(_id)
            return self.find_by_id(_id)
        except ValueError:
            raise InvalidIdException('Incorrect ID.')

    async def select_resource(self):
        base_prompt = BasePrompt()
        resource = self._validate_id(await base_prompt.current_prompt(
            "Select a {} id: ".format(self.resource_type_name),
            toolbar="Enter a {} id.".format(self.resource_type_name),
            autoreturn=True,
            autocomplete=self.id_suggestions
        )
                                     )
        return resource

    async def get_resource(self):
        done = print_waiting_done(
            'getting {}s'.format(self.resource_type_name))
        try:

            self.resources = await self.api_entry_point.get_instances()
            self._populate_id_suggestions()
        except PvApiConnectionError as err:
            print(err)
        finally:
            await done()


class HubCache:
    def __init__(self, request, loop=None):
        self.shades = ResourceCache(PvShades(request), 'shade', request)
        self.rooms = ResourceCache(PvRooms(request), 'room', request)
        self.scenes = ResourceCache(PvScenes(request), 'scene', request)
        self.scene_members = ResourceCache(
            PvSceneMembers(request), 'scene member', request)
        self.loop = loop or get_loop()

    async def update(self):
        await self.shades.get_resource()
        await self.rooms.get_resource()
        await self.scenes.get_resource()
        await self.scene_members.get_resource()
        await asyncio.sleep(1)



import logging
from argparse import ArgumentParser

from aiopvapi.helpers.aiorequest import AioRequest
from aiopvapi.resources.scene import Scene as PvScene
from aiopvapi.resources.shade import BaseShade
from aiopvapi.scene_members import SceneMembers
from nmb.NetBIOS import NetBIOS
from prompt_toolkit.completion import WordCompleter

from prompt_toolkit.eventloop.defaults import use_asyncio_event_loop

from pv_prompt.base_prompts import BasePrompt, PvPrompt, PvResourcePrompt, \
    InvalidIdException, QuitException, Command
from pv_prompt.helpers import get_loop
from pv_prompt.print_output import info, print_scenes, warn, print_shade_data, \
    print_resource_data, print_key_values, print_waiting_done, print_table
from pv_prompt.resource_cache import HubCache
from pv_prompt.zero_conf import Zero

NETBIOS_HUB2_NAME = 'PowerView-Hub'
# todo: get the proper hub1 netbios name.
NETBIOS_HUB1_NAME = 'hub'

LOGGER = logging.getLogger(__name__)


# class ShadeResourceCache(ResourceCache):
#
#     def print_resource_data(self):
#         print("")
#         print_table('NAME', 'ID', 'TYPE')
#         print_table('----', '--', '----')
#         for _item in self.resources:
#             print_table(_item.name, _item.id, _item.shade_type.description)
#         self.print_length()
#         print("")
#
#
# class SceneMemberResourceCache(ResourceCache):
#     def __init__(self, api_entry_point: ApiEntryPoint,
#                  resource_type_name: str,
#                  request: AioRequest,
#                  shades: ShadeResourceCache,
#                  scenes: ResourceCache):
#         super().__init__(api_entry_point,
#                          resource_type_name,
#                          request)
#         self.shades = shades
#         self.scenes = scenes
#
#     def print_resource_data(self):
#         print("")
#         print_table('SHADE', 'SCENE')
#         for _item in self.resources:
#             _shade = self.shades.get_name_by_id(_item.shade_id)
#             _scene = self.scenes.get_name_by_id(_item.scene_id)
#             print_table(_shade.name, _scene.name)
#         self.print_length()


class Shade(PvResourcePrompt):
    def __init__(self, shade: BaseShade, request, hub_cache):
        super().__init__(shade, request, hub_cache)
        self._prompt = 'shade {} {}: '.format(shade.id, shade.name)
        self.register_commands(
            {'j': Command(function_=self.jog, label='(j)og'),
             'o': Command(function_=self.open),
             'c': Command(function_=self.close),
             'r': Command(function_=self.refresh),
             's': Command(function_=self.stop)})

    async def refresh(self, *args, **kwargs):
        await self.pv_resource.refresh()

    async def jog(self, *args, **kwargs):
        await self.pv_resource.jog()

    async def open(self, *args, **kwargs):
        await self.pv_resource.open()

    async def close(self, *args, **kwargs):
        await self.pv_resource.close()

    async def stop(self, *args, **kwargs):
        await self.pv_resource.stop()


class Scenes(PvPrompt):
    def __init__(self, request, hub_cache: HubCache):
        super().__init__(request, hub_cache)
        self.api_resource = hub_cache.scenes
        self.register_commands(
            {'l': Command(function_=self.list_scenes),
             'a': Command(function_=self.activate_scene),
             's': Command(function_=self.select_scene)})

    async def list_scenes(self, *args, **kwargs):
        info("Getting scenes...")
        print_scenes(self.hub_cache.scenes, self.hub_cache.rooms)

    async def activate_scene(self, *args, **kwargs):
        try:
            _scene = await self.hub_cache.scenes.select_resource()

            await _scene.activate()

        except InvalidIdException as err:
            warn(err)

    async def select_scene(self, *args, **kwargs):
        try:
            pv_scene = await self.hub_cache.scenes.select_resource()
            scene = Scene(pv_scene, self.request, self.hub_cache)
            await scene.current_prompt()
        except InvalidIdException as err:
            warn(err)


class Scene(PvResourcePrompt):
    def __init__(self, scene: PvScene, request, hub_cache):
        super().__init__(scene, request, hub_cache)
        self._prompt = 'scene {} {}:'.format(scene.id, scene.name)
        self.register_commands(
            {'a': Command(function_=self.add_shade_to_scene)})

    async def add_shade_to_scene(self, *args, **kwargs):
        shade = await self.hub_cache.shades.select_resource()
        _position = await shade.get_current_position()
        if _position:
            await (SceneMembers(self.request)).create_scene_member(
                _position, self.api_resource.id, shade.id)
        info('Scene created.')

    async def show_members(self, *args, **kwargs):
        info("getting scene members")
        _scene_members = self.hub_cache.scene_members.re


class Rooms(PvPrompt):
    def __init__(self, request, hub_cache: HubCache):
        super().__init__(request, hub_cache)
        self.api_resource = hub_cache.rooms
        self.register_commands({'l': Command(function_=self.list_rooms)})
        self._prompt = "Rooms: "

    async def list_rooms(self, *args, **kwargs):
        info("getting rooms")
        print_resource_data(self.hub_cache.rooms)


class Shades(PvPrompt):
    def __init__(self, request, hub_cache: HubCache):
        super().__init__(request, hub_cache)
        # self._shades_resource = PvShades(request)
        self.api_resource = hub_cache.shades
        self.register_commands(
            {'l': Command(function_=self._list_shades),
             's': Command(function_=self._select_shade)})
        self._prompt = "Shades: "

    async def _list_shades(self, *args, **kwargs):
        info("getting shades...")
        print_shade_data(self.hub_cache.shades)

    async def _select_shade(self, *args, **kwargs):
        try:
            pv_shade = await self.hub_cache.shades.select_resource()
            shade = Shade(pv_shade, self.request, self.hub_cache)
            await shade.current_prompt()
        except InvalidIdException as err:
            warn(err)


class Discovery(BasePrompt):
    def __init__(self):
        super().__init__()
        self._prompt = "Hub connection: "
        self.register_commands(
            {'d': Command(function_=self._discover),
             'c': Command(function_=self.connect, autoreturn=True)})
        self._ip_suggestions = None
        self._ip_completer = None

    async def _discover(self, *args, **kwargs):
        self._ip_suggestions = []
        zero = Zero(self.loop)
        done = print_waiting_done('Discovering hubs')
        LOGGER.debug("discover command fire")
        await zero.discover()
        LOGGER.debug("discovery done")
        await done()
        if zero.hubs:
            self.print_hub_table()
            for _hub in zero.hubs:
                print_key_values(_hub.name, _hub.ip)
                self._ip_suggestions.append(_hub.ip)
        self._populate_completer()

    async def _discover_hub1(self, *args, **kwargs):
        nb = NetBIOS()
        info('Discovering hubs...')
        LOGGER.debug("discover command fire")
        self._ip_suggestions = nb.queryName(NETBIOS_HUB1_NAME, timeout=5)
        if self._ip_suggestions:
            self.print_hub_table()
            for _hub in self._ip_suggestions:
                print_table('hub1', _hub)
            self._populate_completer()
        else:
            warn("...Discovery timed out")

    def print_hub_table(self):
        print('')
        print_table('NAME', 'IP ADDRESS')
        print_table('----', '----------')

    def _populate_completer(self):
        self._ip_completer = WordCompleter(self._ip_suggestions)

    async def connect(self, *args, **kwargs):
        pr = BasePrompt()
        ip = await pr.current_prompt(
            'Enter ip address: ',
            toolbar='Enter a valid ip address: 129.168.2.3',
            autoreturn=True,
            autocomplete=self._ip_completer
        )
        LOGGER.debug('entered ip is : {}'.format(ip))
        try:
            self.validate_ip(ip)
        except ValueError:
            warn("Invalid ip entered.")
        else:
            LOGGER.debug("returning ip address: {}".format(ip))
            return 'http://{}'.format(ip)

    def validate_ip(self, ip):
        vals = ip.split('.')
        try:
            ints = (int(val) for val in vals)
        except ValueError:
            raise ValueError
        for _int in ints:
            if _int < 0 or _int > 256:
                raise ValueError


class MainMenu(BasePrompt):
    def __init__(self, loop, hub=None):
        self.loop = loop
        super().__init__()
        self.request = None
        self.register_commands({'c': Command(function_=self._connect_to_hub)})
        if hub:
            self._register_hub_commands()
        self._prompt = "PowerView toolkit: "
        self._hub_cache = None

    def _register_hub_commands(self):
        self.register_commands(
            {'s': Command(function_=self.shades),
             'e': Command(function_=self.scenes),
             'r': Command(function_=self.rooms),
             'h': Command(function_=self.hub_cache)})

    async def _connect_to_hub(self, *args, **kwargs):
        discovery = Discovery()
        hub = await discovery.current_prompt()
        LOGGER.debug("received hub ip: {}".format(hub))
        if hub:
            info("Using {} as the PowerView hub ip address.".format(hub))
            self.request = AioRequest(hub, loop=self.loop)
            self._register_hub_commands()
            self._hub_cache = HubCache(self.request)

            async def answer_no(*args, **kwargs):
                LOGGER.debug('No entered.')

            async def answer_yes(*args, **kwargs):
                LOGGER.debug('Yes entered.')
                await self.hub_cache(*args, **kwargs)

            pr = BasePrompt(
                commands={'y': Command(function_=answer_yes, autoreturn=True),
                          'n': Command(function_=answer_no, autoreturn=True)})
            await pr.current_prompt(
                prompt_='Query the hub ? <y/n> ')

    async def hub_cache(self, *args, **kwargs):
        LOGGER.debug('Querying the hub.')
        await self._hub_cache.update()

    async def shades(self, *args, **kwargs):
        shade = Shades(self.request, self._hub_cache)
        await shade.current_prompt()

    async def scenes(self, *args, **kwargs):
        scene = Scenes(self.request, self._hub_cache)
        await scene.current_prompt()

    async def rooms(self, *args, **kwargs):
        rooms = Rooms(self.request, self._hub_cache)
        await rooms.current_prompt()

    def close(self):
        if self.request:
            self.loop.run_until_complete(self.request.websession.close())


# logging.basicConfig(level=logging.DEBUG)


def main():
    use_asyncio_event_loop()

    argparser = ArgumentParser()
    argparser.add_argument(
        '--hubip', help="The ip address of the hub", default=None)
    argparser.add_argument(
        '--loglevel', default=30,
        help="Set a custom loglevel. default s 30 debug is 10", type=int)
    args = argparser.parse_args()
    logging.basicConfig(level=args.loglevel)

    loop = get_loop()
    try:
        _main = MainMenu(loop, args.hubip)
        loop.run_until_complete(_main.current_prompt())
    except QuitException:
        print("closing pv toolkit")
    finally:
        _main.close()



import asyncio
import logging
from asyncio import CancelledError
from typing import TYPE_CHECKING

from prompt_toolkit import print_formatted_text, HTML

LOGGER = logging.getLogger(__name__)

if TYPE_CHECKING:
    from pv_prompt.resource_cache import ResourceCache


def print_key_values(key, value):
    print_formatted_text(
        HTML('  <green>{:<15}</green><orange>{}</orange>'.format(key, value)))


def print_table(*values):
    _columns = len(values)
    # _first = '  <green>{:<15}</green>'
    # _second = '<orange>{:<15}</orange>' * (_columns - 1)
    _full = '  <green>{:<15}</green>' + '<orange>{:<15}</orange>' * (
            _columns - 1)
    print_formatted_text(HTML(_full.format(*values)))


def info(text, **kwargs):
    print_formatted_text(
        HTML('  <green>{}</green>'.format(text)), **kwargs
    )


def spinner():
    print_formatted_text(
        HTML('<orange>.</orange>'), end='')



def warn(text):
    print_formatted_text(
        HTML('  <ansired>{}</ansired>'.format(text)))


def print_waiting_done(action):
    async def waiting():
        info(action, end='')
        try:
            while True:
                await asyncio.sleep(0.5)
                LOGGER.debug("spinning")
                spinner()
        except CancelledError:
            pass


    task = asyncio.ensure_future(waiting())

    async def _finished():
        info('done')
        # todo: make the wile loop not infinite as it might hang indefinitely
        task.cancel()
        while not task.cancelled:
            await asyncio.sleep(0.5)

    return _finished


def print_scenes(scenes: 'ResourceCache', rooms):
    print("")
    LOGGER.debug("printing scenes")
    print_table('NAME', 'ID', 'ROOM')
    print_table('----', '--', '----')
    tbl = [
        (_item.name, _item.id,
         rooms.get_name_by_id(_item.room_id)) for _item in
        scenes.resources
    ]
    tbl.sort(key=lambda x: x[-1])

    for _ln in tbl:
        print_table(*_ln)
    print_length(len(scenes.resources))


def print_shade_data(shades):
    print("")
    print_table('NAME', 'ID', 'TYPE')
    print_table('----', '--', '----')
    for _item in shades.resources:
        print_table(_item.name, _item.id, _item.shade_type.description)
    print_length(len(shades.resources))


# def print_resource_data(scene_members: ResourceCache, shades: ResourceCache,
#                         scenes: ResourceCache):


# print("")
#         print_table('SHADE', 'SCENE')
#         for _item in self.resources:
#             _shade = self.shades.get_name_by_id(_item.shade_id)
#             _scene = self.scenes.get_name_by_id(_item.scene_id)
#             print_table(_shade.name, _scene.name)
#         self.print_length()

def print_resource_data(resource):
    print("")
    print_table('NAME', 'ID')
    print_table('----', '--')
    for _item in resource.resources:
        print_table(_item.name, _item.id)
    print_length(len(resource.resources))


def print_length(length):
    print('')
    info('{} items found.'.format(length))
    print('')



import asyncio
import json
import logging
from typing import TYPE_CHECKING

from aiopvapi.helpers.aiorequest import AioRequest
from aiopvapi.helpers.api_base import ApiResource
# from prompt_toolkit import prompt
from prompt_toolkit import PromptSession

from pv_prompt.print_output import info

LOGGER = logging.getLogger(__name__)

if TYPE_CHECKING:
    from pv_prompt.resource_cache import HubCache


class QuitException(Exception):
    pass


class BackException(Exception):
    pass


class InvalidIdException(Exception):
    pass


class Command:
    """Command class. Is bound to a key press."""

    def __init__(self, function_=None, label=None,
                 autoreturn=False):
        self._function = function_ or self._empty
        # if not asyncio.iscoroutine(self._function):
        #     raise TypeError("function is not a coroutine: {}".format(
        #         self._function.__name__))
        self.autoreturn = autoreturn
        self._label = label

    def __call__(self, *args, **kwargs):
        return self._function(*args, **kwargs)

    async def _empty(self, *args, **kwargs):
        return None

    @property
    def name(self):
        if self._label:
            return self._label
        else:
            return self._function.__name__

    def __repr__(self):
        return '{}({},{},{}'.format(
            self.__class__.__name__,
            self._function.__name__,
            self.name,
            self.autoreturn)


class BasePrompt:
    def __init__(self, commands=None, loop=None):
        if loop:
            self.loop = loop
        else:
            self.loop = asyncio.get_event_loop()
        self._prompt = "Enter a command: "
        self._commands = {'q': Command(function_=self.quit, label='(q)uit'),
                          'b': Command(autoreturn=True, label='back')}
        if commands:
            self.register_commands(commands)
        self._auto_return = False

    @property
    def commands(self):
        return self._commands

    def _toolbar_string(self):
        _str = ' | '.join(
            ('({}) {}'.format(key, value.name) for key, value in
             self._commands.items()))
        return _str

    def register_commands(self, commands: dict):
        """Adds commands key strokes to watch in that context"""
        self._commands.update(commands)

    async def current_prompt(self, prompt_=None, toolbar=None,
                             autocomplete=None,
                             autoreturn=False):
        """The currently active prompt.

        """
        self._auto_return = autoreturn
        if toolbar is not None:
            self._toolbar_string = lambda: toolbar
        if prompt_ is None:
            prompt_ = self._prompt
        try:
            while True:
                prompt = PromptSession(prompt_, bottom_toolbar=self._toolbar_string,
                                completer=autocomplete)
                _command = await prompt.prompt(async_=True)
                LOGGER.debug("received command: {}".format(_command))
                _meth = self.commands.get(_command)
                if _meth:
                    LOGGER.debug(
                        "a method tied to this command: {}".format(_meth))
                    val = await _meth(_command)
                    LOGGER.debug("method return with value: {}".format(val))
                    if _meth.autoreturn:
                        LOGGER.debug("auto returning from the method")
                        return val
                if self._auto_return:
                    LOGGER.debug("Returning with entered command.")
                    return _command
        except BackException:
            return None

    # def back(self, *args, **kwargs):
    #     raise BackException()

    async def quit(self, *args, **kwargs):
        raise QuitException()


class PvPrompt(BasePrompt):
    # def __init__(self, request: AioRequest, commands=None):
    def __init__(self, request: AioRequest, hub_cache: 'HubCache',
                 commands=None):
        super().__init__(commands=commands)
        self.hub_cache = hub_cache
        self.request = request

        if request:
            self.hub_ip = request.hub_ip
        else:
            self.hub_ip = 'not connected'
        self.api_resource = None

    def _toolbar_string(self):
        line1 = super()._toolbar_string()
        return line1 + ' \nhub: {}'.format(self.hub_ip)


class PvResourcePrompt(PvPrompt):
    def __init__(self, pv_resource: ApiResource, request: AioRequest,
                 hub_cache: 'HubCache'):
        super().__init__(request, hub_cache)
        self.pv_resource = pv_resource
        self.register_commands(
            {'s': Command(function_=self.show_raw, label='(s)how raw')})

    async def show_raw(self, *args, **kwargs):
        _str = json.dumps(self.pv_resource.raw_data, indent=4)
        info(_str)
